"""
Helper functions for accessing the data from other databases from within
IPython notebooks.
"""
import requests
import pandas
from io import StringIO
from pandas import read_csv
import numpy as np
from IPython.display import Image


def genes_of_interest_from_string(gene_names,
                                  no_of_interacting_partners, taxonomyid):
    url = 'https://string-db.org/api/psi-mi-tab/interactionsList?identifiers='
    for g in gene_names:
        url = url + g + '%0d'
    url = (url + '&limit=' + str(no_of_interacting_partners) +
           '&network_flavor=evidence&species=' + taxonomyid)
    res = requests.get(url)
    df = read_csv(StringIO(res.text), sep='\t', header=None)
    c1 = df.loc[:, [2, 3, 14]]
    return c1


def network_of_interest(gene_names, no_of_interacting_partners):
    url = 'https://string-db.org/api/image/network?identifier='
    limit = str(no_of_interacting_partners) + '&network_flavor=evidence'
    for g in gene_names:
        url = url + g  # This doesn't look like it will work

    url = url + '&limit=' + limit
    return Image(url=url)


def get_entrezid(gene):
    """
    Use mygene.info api to convert gene_symbols to
    entrezgeneIds (gets ortholog ids as well)

    Note : This returns entrez ids for the gene in all organisms
    (Could consider taxonomy related entrez ids : if need be!)
    documentation for api:

    http://mygene.info/v3/api/#MyGene.info-gene-query-service-GET-Gene-query-service
    """
    entrezurl = "https://mygene.info/v3/query?q="
    entrezurl = entrezurl+gene

    res = requests.get(entrezurl)
    if not res.ok:
        return []

    results = pandas.read_json(StringIO(res.text))

    entrezid = []
    if results.empty:
        return entrezid

    for i in results.hits:
        key = list(i.keys())
        value = list(i.values())
        for cntr, k in enumerate(key):
            if k == 'entrezgene':
                entrezid.append(value[cntr])
                return entrezid


def get_ensembleid(gene):

    ensembleserver = "https://rest.ensembl.org/xrefs/symbol/homo_sapiens/"
    url = ensembleserver + gene + "?content-type=application/json"

    res = requests.get(url)
    results = pandas.read_json(StringIO(res.text))

    ensembleid = []
    if results.empty:
        return ensembleid

    for i in results.iloc[:, 0]:
        ensembleid.append(i)
    return ensembleid


def genes_of_interest_go(go_term, taxonomy_id):

    go = "https://www.ebi.ac.uk/QuickGO/services/annotation/downloadSearch"
    url = (go + '?taxonId=' + taxonomy_id + '&goId=%20' + go_term +
           '%20&format=tsv')
    res = requests.get(url, headers={"Accept": "text/tsv"})
    df = read_csv(StringIO(res.text), sep='\t', header=None)
    c1 = df.iloc[:, 2]
    return list(set(np.unique(c1.values.ravel())) - set(['Symbol', '-']))


def ensembleid_to_genesymbol(ensembleid):

    ensembleserver = "https://rest.ensembl.org/xrefs/id/"
    url = (ensembleserver + ensembleid +
           "?content-type=application/json;external_db=WikiGene")
    res = requests.get(url)
    if not res.ok:
        return ensembleid
    if "error" in res.text:
        return ensembleid
    results = pandas.read_json(StringIO(res.text))
    if results.empty:
        return ensembleid
    symbol = results['display_id'][0]
    return symbol



global organism
global idr_base_url
global hostname
global username
global taxonomy_id
global ensemble_id
global gene_symbol
organism = 'Homo Sapiens'
taxonomy_id = '9606'
ensemble_id = 'ENSG00000106526'
gene_symbol = 'ACTR3C'
go_gene_list = ['SOD1', 'SOD2', 'SO']



"""
Helper functions for accessing the IDR from within IPython notebooks.
"""
import pandas
import omero.clients # NOQA
from omero.rtypes import rlist, rstring, unwrap
from omero.sys import ParametersI
from .externalDBs import get_entrezid, get_ensembleid, ensembleid_to_genesymbol

from .widgets import progress
import numpy as np


def attributes_by_attributes(conn,
                             name="Gene Symbol",
                             value="ASH2L",
                             ns="openmicroscopy.org/mapr/gene",
                             ns2="openmicroscopy.org/mapr/phenotype",
                             name2=None,
                             s_id=None
                             ):

    """
    Return a list of neighbours attributes
    for given case insensitive attribute value. (Uses the python blitz gateway)
    """
    params = ParametersI()
    params.addString("value", value.lower())
    q = (
        "select distinct new map( mv.value as value) "
        "from Annotation as a "
        "join a.mapValue as mv "
        "where lower(mv.value) = :value {where_claus}"
    )
    where_claus = []
    if name:
        params.addString("name", name)
        where_claus.append("and mv.name = :name")
    q = q.format(**{'where_claus': " ".join(where_claus)})

    values = [v[0]['value'] for v in unwrap(
        conn.getQueryService().projection(q, params))]

    params = ParametersI()
    valuelist = [rstring(str(v)) for v in values]
    params.add('values', rlist(valuelist))
    params.addString("ns", ns)
    params.addString("ns2", ns2)

    q = (
        "select distinct new map("
        "mv.name as name, "
        "mv.value as value, "
        "mv2.name as name2, "
        "mv2.value as value2) "
        "from Image as i "
        "join i.annotationLinks as ial "
        "join i.annotationLinks as ial2 "
        "join ial.child as a "
        "join a.mapValue as mv "
        "join ial2.child as a2 "
        "join a2.mapValue as mv2 "
        "where a.ns = :ns and a2.ns = :ns2 "
        "and mv.value in (:values) {where_claus} "
    )

    where_claus = []
    if name:
        params.addString("name", name)
        where_claus.append("and mv.name = :name")
    if name2:
        params.addString("name2", name2)
        where_claus.append("and mv2.name = :name2")

    q = q.format(**{'where_claus': " ".join(where_claus)})

    if s_id is not None:
        q = q + ("and i in (select image from WellSample "
                 "where well.plate in "
                 "(select child from ScreenPlateLink where "
                 "parent.id = {sId}))")

        screen_id_list = []
        screen_id_list.append(str(s_id))
        q = q.format(**{'sId': " ".join(screen_id_list)})

    res = {}
    for r in unwrap(conn.getQueryService().projection(q, params)):
        r = r[0]
        try:
            res[(r['name'], r['value'])].append((r['name2'], r['value2']))
        except KeyError:
            res[(r['name'], r['value'])] = [(r['name2'], r['value2'])]
    return res


def annotation_ids_by_field(conn,
                            value="CMPO_0000077",
                            key="Phenotype Term Accession",
                            ns="openmicroscopy.org/mapr/phenotype"):
    """
    Return a list of IDs for map annotations with the given namespace
    that have a key=value pair matching the given parameters.
    """

    params = ParametersI()
    params.addString("value", value)
    params.addString("key", key)
    params.addString("ns", ns)
    q = ("select a.id from MapAnnotation a join a.mapValue as mv "
         "where a.ns = :ns and mv.name = :key and mv.value = :value")

    return unwrap(conn.getQueryService().projection(q, params))[0]


def get_phenotypes_for_gene(session,
                            gene_name,
                            screenid=None,
                            idr_base_url="https://idr.openmicroscopy.org"):

    """
    Return a list of phenotype
    for given case insensitive gene_name. (Uses the json api)
    """

    v = "{base}/mapr/api/{key}/"
    screens_projects_url = v + "?value={value}"
    plates_url = v + "plates/?value={value}&id={screen_id}"
    images_url = v + "images/?value={value}&node={parent_type}&id={parent_id}"
    map_url = "{base}/webclient/api/annotations/?type=map&{type}={image_id}"

    screen_id_list = []
    if screenid is not None:
        screen_id_list.append(screenid)
    else:
        qs = {'base': idr_base_url, 'key': 'gene', 'value': gene_name}
        url = screens_projects_url.format(**qs)
        for s in session.get(url).json()['screens']:
            screen_id_list.append(s['id'])

    unique_list = []
    unique_list_1 = []
    unique_list_2 = []

    if len(screen_id_list) == 0:
        phenotype_ids_dataframe = pandas.DataFrame(
            {'Name': unique_list,
             'Accession': unique_list_1,
             'phenotypeAndScreenId': unique_list_1})
        return phenotype_ids_dataframe

    for sid in screen_id_list:
        screen_id = sid
        screenqs = {'base': idr_base_url, 'key': 'gene',
                    'value': gene_name, 'screen_id': screen_id}
        screenurl = plates_url.format(**screenqs)
        phenotype_per_screen = []
        phenotype_id_per_screen = []
        screen_ids = []
        t_name = 'Phenotype Term Name'
        t_access = 'Phenotype Term Accession'
        for p in session.get(screenurl).json()['plates']:
            plate_id = p['id']
            imageqs = {'base': idr_base_url, 'key': 'gene',
                       'value': gene_name, 'parent_type': 'plate',
                       'parent_id': plate_id}
            plateurl = images_url.format(**imageqs)
            for i in session.get(plateurl).json()['images']:
                image_id = i['id']
                qs = {'base': idr_base_url,
                      'type': 'image',
                      'image_id': image_id}
                url = map_url.format(**qs)
                for a in session.get(url).json()['annotations']:
                    for v in a['values']:
                        key = v[0]
                        value = v[1]
                        if key.startswith(t_name) & key.endswith(t_name):
                            phenotype_per_screen.append(value)
                            screen_ids.append(str(screen_id) + '_' + value)
                        if key.startswith(t_access) & key.endswith(t_access):
                            phenotype_id_per_screen.append(value)

        unique_list = unique_list + list(set(phenotype_per_screen))
        unique_list_1 = unique_list_1 + list(set(phenotype_id_per_screen))
        unique_list_2 = unique_list_2 + list(set(screen_ids))

        phenotype_ids_dataframe = pandas.DataFrame(
            {'Name': unique_list,
             'Accession': unique_list_1,
             'phenotypeAndScreenId': unique_list_2})

    return phenotype_ids_dataframe


def get_phenotypes_for_genelist(session,
                                go_gene_list,
                                organism,
                                idr_base_url="https://idr.openmicroscopy.org",
                                lookup_entrez=True):

    """
    Return a list of phenotypes (dataframe)
    for given case insensitive gene_list. (Uses the json api)
    """

    genedict = {}
    totalphenotypename = []
    totalphenotypeaccession = []
    totalscreenids = []
    testedgenes = []

    phenotypenametoacc = {}
    for ids, gene in enumerate(go_gene_list):

        if gene.startswith("-"):
            continue

        if lookup_entrez:
            entrezid = get_entrezid(gene)
        else:
            entrezid = '-'
        ensembleid = get_ensembleid(gene)

        gid = None
        # search with the gene name
        uniquelist = []
        if len(uniquelist) == 0:
            key = "GeneName"
            gid = gene
            uniquelist = get_phenotypes_for_gene(session,
                                                 gid)

        # search with ensembleid if geneSymbol does not
        # return any result
        if len(uniquelist['Name']) == 0:
            key = "EnsemblID"
            for gid in ensembleid:
                uniquelist = get_phenotypes_for_gene(session,
                                                     gid)
                if len(uniquelist['Name']) != 0:
                    break

        # search with entrezid if gene symbol and
        # ensembleid does not return any result
        if lookup_entrez:
            if len(uniquelist) == 0:
                key = "EntrezID"
                for gid in entrezid:
                    uniquelist = get_phenotypes_for_gene(session,
                                                         gid)
                    if len(uniquelist['Name']) != 0:
                        break

        # List of genes from string which were part of IDR
        if gid is not None:
            testedgenes.append(gid)

        # Dataframe of genes from string which were part of IDR
        # and had a phenotype associated with them
        if len(uniquelist) != 0:

            accname = uniquelist['Name']
            accid = uniquelist['Accession']
            scrid = uniquelist['phenotypeAndScreenId']

            accnames = list(accname.values)
            accids = list(accid.values)
            idlist = []
            for id in scrid:
                idx = id.index('_')
                idlist.append(id[:idx])

            for idx, idx1 in enumerate(accnames):
                phenotypenametoacc[accnames[idx]] = accids[idx]

            totalphenotypename = totalphenotypename + accnames
            totalphenotypeaccession = totalphenotypeaccession + accids
            totalscreenids = totalscreenids + list(scrid.values)

            genedict[gene] = [entrezid, ensembleid,
                              None, None, None, None, None]
            genedict[gene][2] = key
            genedict[gene][3] = gid
            genedict[gene][4] = accnames
            genedict[gene][5] = accids
            genedict[gene][6] = idlist

        progress(ids+1, len(go_gene_list),
                 status='Iterating through gene list')

    query_genes_dataframe = pandas.DataFrame.from_dict(genedict,
                                                       orient='index')
    query_genes_dataframe.columns = ("Entrez", "Ensembl",
                                     "Key", "Value", "PhenotypeName",
                                     "PhenotypeAccession", "ScreenIds")

    # get the screens to phenotypes map for the query genes
    organism_screen_idlist = get_organism_screenids(session, organism)
    genes_scid_list = [item for sublist in
                       query_genes_dataframe['ScreenIds'].values
                       for item in sublist]
    genes_scid_list = list(set(genes_scid_list))
    screen_to_phenotype_dictionary = {}
    for scid in genes_scid_list:
        if scid in organism_screen_idlist:
            content = [x for x in set(list(totalscreenids))
                       if x.startswith(scid)]
            for idx, item in enumerate(content):
                idx1 = item.index('_')
                content[idx] = item[idx1+1:]
            screen_to_phenotype_dictionary[scid] = content

    return [query_genes_dataframe, screen_to_phenotype_dictionary]


def get_similar_genes(conn, query_genes_list, screen_to_phenotype_dictionary):

    """
    Return a multi-dimensional dictionary with the
    following mapping,
    similar_genes[screenid][phenotypename] = similar_genes_list
    (uses python blitz gateway)
    """

    similar_genes = {}
    overlap_genes = {}
    scid_list = set(list(screen_to_phenotype_dictionary.keys()))
    for i, sid in enumerate(set(scid_list)):

        phlist = screen_to_phenotype_dictionary[sid]
        similar_genes[str(sid)] = {}
        overlap_genes[str(sid)] = {}
        for phenotype in np.unique(phlist):

            args = {
                "name": "Phenotype Term Name",
                "value": phenotype,
                "ns": "openmicroscopy.org/mapr/phenotype",
                "ns2": "openmicroscopy.org/mapr/gene",
                "s_id": sid
            }

            cc = attributes_by_attributes(conn, **args)
            dataframe = pandas.DataFrame.from_dict(cc)

            if dataframe.empty:
                continue

            gene_list = []
            for x in dataframe.iloc[:, 0]:

                key = x[0]
                value = x[1]

                if key == "Gene Identifier" and value.startswith("EN"):
                    id = value
                    gene_list.append(id)
                if key == "Gene Symbol":
                    genesym = value
                    gene_list.append(genesym)

            ov_genes = set(gene_list).intersection(query_genes_list)

            remove_duplicates = []
            removed_genes = []
            for g in ov_genes:
                converted = g
                if g.startswith('ENSG'):
                    converted = ensembleid_to_genesymbol(g)
                    removed_genes.append(g)
                remove_duplicates.append(converted)

            ov_genes = set(remove_duplicates)

            if len(ov_genes) != 0:
                setdiff_genes = set(gene_list) - ov_genes - set(removed_genes)
                similar_genes[str(sid)][phenotype] = list(setdiff_genes)
                overlap_genes[str(sid)][phenotype] = list(ov_genes)

        progress(i+1, len(set(scid_list)), status='Iterating through screens')

    return [similar_genes, overlap_genes]


def get_organism_screenids(session, organism,
                           idr_base_url="https://idr.openmicroscopy.org"):

    """
    Return a list of screen ids in IDR
    for given case insensitive organism. (Uses the json api)
    """

    screen_id_list = []
    qs = {'base': idr_base_url, 'key': 'organism', 'value': organism}
    screens_projects_url = "{base}/mapr/api/{key}/?value={value}"
    url = screens_projects_url.format(**qs)
    sr = session.get(url)
    sr.raise_for_status()
    for s in sr.json()['screens']:
        screen_id_list.append(str(s['id']))

    return screen_id_list



"""
Helper functions for accessing the IDR from within IPython notebooks.
"""
from scipy.stats import pearsonr
import matplotlib.pyplot as plt
from .attributes import annotation_ids_by_field


def images_by_phenotype(conn, phenotype="CMPO_0000077"):
    """
    Passes phenotype as the value argument to annotation_ids_by_field
    and loads Image objects which can be used for loading thumbnails, etc.
    """
    ann_ids = annotation_ids_by_field(conn, phenotype)
    return list(conn.getObjectsByAnnotations("Image", ann_ids))


def simple_colocalisation(image):
    """
    Perform a simple comparison of the red and green channels.

    Pearson coefficient described at
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3074624/
    """
    p = image.getPrimaryPixels()

    # get 2D planes and reshape to 1D array
    r = p.getPlane(0, 0, 0)
    red = r.reshape(r.size)
    g = p.getPlane(0, 1, 0)
    green = g.reshape(g.size)

    # pearson colocalistion coefficient
    pearsonr(red, green)

    # scatter plot
    plt.scatter(red, green)
    plt.show()



import sys
from ipywidgets import widgets
from IPython.display import display


def dropdown_widget(organisms_list,
                    dropdown_widget_name,
                    displaywidget=False):

    org_sel = widgets.Dropdown(
        options=organisms_list,
        value=organisms_list[0],
        description=dropdown_widget_name,
        disabled=False,
    )
    if displaywidget is True:
        display(org_sel)
    return org_sel


def textbox_widget(temp_value,
                   tool_tip,
                   textbox_name,
                   displaywidget=False):

    text_box = widgets.Text(
        value=temp_value,
        placeholder=tool_tip,
        description=textbox_name,
        disabled=False
    )

    if displaywidget is True:
        display(text_box)
    return text_box


def html_widget(text, displaywidget=False):

    htmlwidget = widgets.HTML(
        value="<b>" + text + "</b>",
        placeholder='Some HTML',
        description='',
    )
    if displaywidget is True:
        display(htmlwidget)
    return htmlwidget


def progress(count, total, status=''):

    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', status))
    # As suggested by Rom Ruben
    # (see: http://stackoverflow.com/
    # questions/3173320/text-progress-bar-in-the-console/
    # 27871113#comment50529068_27871113)
    sys.stdout.flush()



from .connections import connection # NOQA
from .connections import create_http_session # NOQA

from .attributes import attributes_by_attributes # NOQA
from .attributes import annotation_ids_by_field # NOQA
from .attributes import get_phenotypes_for_gene # NOQA
from .attributes import get_phenotypes_for_genelist # NOQA
from .attributes import get_similar_genes # NOQA
from .attributes import get_organism_screenids # NOQA

from .images import images_by_phenotype # NOQA
from .images import simple_colocalisation # NOQA

from .externalDBs import genes_of_interest_from_string # NOQA
from .externalDBs import network_of_interest # NOQA
from .externalDBs import get_entrezid # NOQA
from .externalDBs import get_ensembleid # NOQA
from .externalDBs import genes_of_interest_go # NOQA
from .externalDBs import ensembleid_to_genesymbol # NOQA

from .visualizations import plot_idr_attributes # NOQA
from .visualizations import plot_string_interactions # NOQA



from collections import Counter
from collections import OrderedDict
import matplotlib.pyplot as plt
import pandas
import seaborn as sns
from .externalDBs import ensembleid_to_genesymbol


def plot_idr_attributes(primary_dictionary,
                        secondary_dictionary,
                        plot_title,
                        filter_by_category,
                        threshold_for_category,
                        threshold_for_plot,
                        plot_condition=True):
    gene_counts = []
    screenids_removed = []
    phenotypes_removed = []
    genes_of_interest = []
    query_list = []
    for screenid in list(primary_dictionary.keys()):
        phenolist = []
        screens_list = []
        query_list = []
        for phenoid in list(primary_dictionary[screenid].keys()):
            phenolist = primary_dictionary[screenid][phenoid]
            phenolist1 = secondary_dictionary[screenid][phenoid]
            if filter_by_category == 'Phenotypes':
                if len(phenolist1) < threshold_for_category:
                    phenolist = []
                    phenotypes_removed.append(phenoid)
            screens_list = screens_list + phenolist
            query_list = query_list + phenolist1

        if filter_by_category == 'Screens':
            screens_list = list(set(screens_list))
        if len(screens_list) >= threshold_for_category:
            gene_counts = gene_counts + screens_list
        else:
            screenids_removed.append(screenid)
    letter_counts = Counter(gene_counts)
    keystoremove = []
    for key, value in letter_counts.items():
        if value < threshold_for_plot:
            keystoremove.append(key)
    for keys in keystoremove:
        del letter_counts[keys]
    query_list = list(set(query_list))
    letter_counts = OrderedDict(letter_counts.most_common())

    if primary_dictionary != secondary_dictionary:
        dict1 = OrderedDict()
        for k, v in letter_counts.items():
            key = k
            if k.startswith('ENSG'):
                key = ensembleid_to_genesymbol(k)
                if key in dict1:
                    value1 = dict1[key]
                    if value1 > v:
                        v = value1
            if key in list(query_list):
                print('in query list')
                continue
            dict1[key] = v
        letter_counts = dict1

    if '' in letter_counts:
        del letter_counts['']

    genes_of_interest = list(letter_counts.keys())

    df = pandas.DataFrame.from_dict(letter_counts, orient='index')

    if plot_condition is True:
        if df.empty:
            print('DataFrame is empty, please reduce thresholds!')
            hf, ha = plt.subplots(1, 1)
        else:
            ax = df.plot(kind='bar', figsize=(30, 15), fontsize=18)
            ax.set_title(plot_title, fontsize=18)
            ax.set_xlabel("Genes", fontsize=18)
            ax.set_ylabel("Number of Unique"
                          + filter_by_category + "in IDR", fontsize=18)
        plt.show()
    else:
        print("Plots are currently" +
              "shown in the notebooks alone"
              " and not from the terminal")

    return screenids_removed, phenotypes_removed, genes_of_interest


def plot_string_interactions(primary_list,
                             secondary_list,
                             total_interactions_dataframe,
                             plot_condition=True):
    dict1 = {}
    for gene in primary_list:
        c2 = total_interactions_dataframe.loc[total_interactions_dataframe[2]
                                              == gene]
        c3 = total_interactions_dataframe.loc[total_interactions_dataframe[3]
                                              == gene]
        c4 = pandas.concat([c2, c3])

        dict1[str(gene)] = {}
        totlist = set(list(c2[3]) + list(c3[2]))
        intwithsublist = totlist.intersection(secondary_list)
        if len(intwithsublist) > 0:
            for gene1 in intwithsublist:
                c5 = c4.loc[c4[2] == gene1]
                if c5.empty:
                    c5 = c4.loc[c4[3] == gene1]
                score = str(c5[14])
                stid = score.index('score:')
                endid = score.index('|')
                dict1[str(gene)][gene1] = float(score[stid+6:endid])

    df = pandas.DataFrame.from_dict(dict1, orient='index')
    df = df.fillna(value=int(0))
    df['ColTotal'] = df.sum(axis=1)
    df.loc['RowTotal'] = df.sum()
    df = df.sort_values(by='RowTotal', ascending=False, axis=1)
    df = df.sort_values(by='ColTotal', ascending=False)
    df = df.drop(['RowTotal'])
    df = df.drop(['ColTotal'], axis=1)

    if plot_condition is True:
        if len(df.columns) == 1 or len(df.index) == 1:
            sns.heatmap(df)
            plt.show()
        elif df.empty:
            print('No primary interactors found')
        else:
            g = sns.clustermap(df)
            plt.setp(g.ax_heatmap.yaxis.get_majorticklabels(), rotation=0)
            plt.show()
    else:
        print("Plots are currently" +
              "shown in the notebooks alone"
              " and not from the terminal")
    return df



"""
Helper functions for accessing the IDR from within IPython notebooks.
"""
import re
import requests
import os
import sys

import omero
from omero.gateway import BlitzGateway


def _configuration_from_url(config_url):
    """
    OMERO binary protocol doesn't support load balancing nor session pinning
    so it has to be done client-side by connecting to a random server/port
    """
    try:
        host = re.match(r'\w+://([^:/]+)', config_url).group(1)
    except AttributeError:
        host = config_url
        config_url = 'https://%s/connection/omero-client.json' % host
    r = requests.get(config_url)
    r.raise_for_status()
    cfg = r.json()
    return cfg, host


def _lookup_parameter(initial, paramname, default):
    if initial is not None:
        return initial
    v = os.getenv('IDR_' + paramname.upper())
    if v is not None:
        return v
    return default


def _host_includes_ice_proto(host):
    return host and re.match(r'(ws|wss|tcp|ssl)://', host)


def connection(host=None, user=None, password=None, port=None, verbose=1):
    """
    Connect to the IDR analysis OMERO server
    Lookup of connection parameters is done in this order:

    1. If host/IDR_HOST starts with protocol:// and protocol is not an Ice
       transport treat this as a full omero-client.json configuration URL and
       fetch it
    2. If host/IDR_HOST starts with an Ice transport:// connect directly
    3. If host/IDR_HOST is defined but port/IDR_PORT empty and there is no
       protocol then attempt to fetch configuration from
       https://host/connection/omero-client.json
    4. Remaining parameters are taken first from the method arguments,
       then from IDR_{HOST,PORT,USER,PASSWORD}
    5. If host/IDR_HOST was a configuration URL then automatically set
       host to the host portion of the URL in case it needs to be
       substituted into the fetched configuration

    To support firewalled environments if host/IDR_HOST has no protocol and the
    connection fails automatically attempt to reconnect using websockets (wss)
    on port 443.

    No defaults are provided

    :return: A BlitzGateway object
    """
    host = _lookup_parameter(host, 'host', '')
    port = int(_lookup_parameter(port, 'port', 0))
    user = _lookup_parameter(user, 'user', None)
    password = _lookup_parameter(password, 'password', None)

    autocfg = []
    if ((host and not port) or re.match(r'\w+://', host)) and (
            not _host_includes_ice_proto(host)):
        autocfg, host = _configuration_from_url(host)

    # https://github.com/openmicroscopy/openmicroscopy/blob/v5.4.3/components/tools/OmeroPy/src/omero/clients.py#L50
    kwargs = {'args': autocfg}
    if host:
        kwargs['host'] = host
    if port:
        kwargs['port'] = port

    c = omero.client(**kwargs)

    # If omero.user and omero.pass were included in omero-client.json
    # they can optionally be omitted in createSession
    create_session_args = []
    if user:
        create_session_args.append(user)
        if password:
            create_session_args.append(password)

    try:
        c.createSession(*create_session_args)
        c.enableKeepAlive(300)
        conn = BlitzGateway(client_obj=c)
    except omero.ClientError as e:
        if re.match(r'\w+://', host):
            raise
        print('Failed to connect: {}, retrying with websockets'.format(e),
              file=sys.stderr)
        return connection(
            'wss://{}/omero-ws'.format(host), user, password, 443, verbose)

    if verbose > 0:
        server = ''
        if verbose > 1:
            info = conn.c.sf.ice_getConnection().getInfo()
            server = '[{}:{}]'.format(info.remoteAddress, info.remotePort)

        print("Connected to IDR%s ..." % server)
    return conn


def create_http_session(idr_base_url='https://idr.openmicroscopy.org'):

    """
    Create and return http session
    """
    index_page = "%s/webclient/?experimenter=-1" % idr_base_url

    # create http session
    with requests.Session() as session:
        request = requests.Request('GET', index_page)
        prepped = session.prepare_request(request)
        response = session.send(prepped)
        if response.status_code != 200:
            response.raise_for_status()

    return session



import pytest
import pandas as pd
from idr import get_phenotypes_for_genelist
from idr import get_phenotypes_for_gene
from idr import get_similar_genes
from idr import get_organism_screenids
from idr import create_http_session, connection
from idr import images_by_phenotype

from .config import organism


class TestAttributes():

    @classmethod
    def setup_class(cls):
        cls.session = create_http_session()
        cls.conn = connection()

    def test_images_by_phenotype(self):

        images = images_by_phenotype(self.conn)
        assert (isinstance(images, list) is True)

    @pytest.mark.parametrize(
        'organism,empty', [
            pytest.param(
                'Homo sapiens', False,
                marks=pytest.mark.xfail(reason="may timeout")),
            ('Saccharomyces cerevisiae', False),
            ('Schizosaccharomyces pombe', False),
            ('Drosophila melanogaster', False),
            ('Mus musculus', True),
            ('Arabidopsis thaliana', True),
        ])
    def test_get_organism_screenids(self, organism, empty):
        organism_screen_idlist = get_organism_screenids(
            self.session, organism)
        if empty:
            assert (organism_screen_idlist == [])
        else:
            assert (organism_screen_idlist != [])

    def test_get_phenotypes_for_genelist(self):

        gene_list = ['CCS', 'SOD2', 'SOD3', 'SOD1']
        [query_genes_df,
         screen_to_phenotype_dict] = get_phenotypes_for_genelist(self.session,
                                                                 gene_list,
                                                                 organism)

        assert isinstance(query_genes_df, pd.DataFrame)
        assert isinstance(screen_to_phenotype_dict, dict)
        assert not query_genes_df.empty
        print(screen_to_phenotype_dict)
        assert not bool(screen_to_phenotype_dict)

        [similar_genes,
         overlap_genes] = get_similar_genes(self.conn,
                                            gene_list,
                                            screen_to_phenotype_dict)

        assert isinstance(similar_genes, dict)
        assert isinstance(overlap_genes, dict)
        assert (similar_genes != [])
        assert (overlap_genes != [])

    def test_get_phenotypes_for_gene(self):

        gid = 'ARPC2'
        sid = 206
        uniquelist = get_phenotypes_for_gene(self.session, gid)

        assert isinstance(uniquelist, pd.DataFrame)
        assert not uniquelist.empty
        uniquelist = get_phenotypes_for_gene(self.session, gid, sid)

        assert isinstance(uniquelist, pd.DataFrame)
        assert not uniquelist.empty

    @classmethod
    def teardown_class(cls):
        cls.conn.close()



import pandas as pd
from idr import genes_of_interest_from_string
from idr import create_http_session, connection
from idr import get_phenotypes_for_genelist
from idr import get_similar_genes
from idr import plot_idr_attributes, plot_string_interactions

from .config import go_gene_list, taxonomy_id, organism


class TestVisualizations():

    @classmethod
    def setup_class(cls):
        cls.session = create_http_session()
        cls.conn = connection()

    def test_plot_string_interactions(self):

        intpartners = genes_of_interest_from_string(go_gene_list,
                                                    20,
                                                    taxonomy_id)
        assert isinstance(intpartners, pd.DataFrame)
        assert not intpartners.empty

        df = plot_string_interactions(go_gene_list,
                                      go_gene_list,
                                      intpartners,
                                      False)

        assert isinstance(df, pd.DataFrame)
        assert not df.empty

    def test_plot_idr_attributes(self):

        [query_genes_df,
         screen_to_phenotype_dict] = get_phenotypes_for_genelist(self.session,
                                                                 go_gene_list,
                                                                 organism)
        query_genes_list = list(query_genes_df['Value'])
        [similar_genes,
         overlap_genes] = get_similar_genes(self.conn,
                                            query_genes_list,
                                            screen_to_phenotype_dict)

        [screenids_removed,
         phenotypes_removed,
         genes_of_interest] = plot_idr_attributes(similar_genes,
                                                  overlap_genes,
                                                  'int_test',
                                                  'Phenotypes',
                                                  0, 5,
                                                  False)

        assert isinstance(screenids_removed, list)
        assert isinstance(phenotypes_removed, list)
        assert isinstance(genes_of_interest, list)

        [screenids_removed,
         phenotypes_removed,
         genes_of_interest] = plot_idr_attributes(overlap_genes,
                                                  similar_genes,
                                                  'int_test',
                                                  'Screens',
                                                  0, 0,
                                                  False)

        assert isinstance(screenids_removed, list)
        assert isinstance(phenotypes_removed, list)
        assert isinstance(genes_of_interest, list)

    @classmethod
    def teardown_class(cls):
        cls.conn.close()



from idr.widgets import dropdown_widget, textbox_widget, html_widget, progress


class TestWidgets():

    def test_textbox_widget(self):
        idr_username = textbox_widget('int_test',
                                      'Enter Username',
                                      'IDR username:',
                                      False)

        assert idr_username.value == 'int_test'
        assert idr_username.description == 'IDR username:'
        assert idr_username.placeholder == 'Enter Username'

    def test_dropdown_widget(self):

        organisms_list = ['A', 'B', 'C']
        org_sel = dropdown_widget(organisms_list,
                                  'Select Organism:',
                                  False)

        assert org_sel.value == 'A'
        assert org_sel.description == 'Select Organism:'

    def test_html_widget(self):

        text = 'int_test'
        htmlwid = html_widget(text, displaywidget=False)

        assert htmlwid.value == '<b>' + text + '</b>'

    def test_progress(self):

        scid_list = ['1', '2', '3']
        for i, sid in enumerate(set(scid_list)):
            progress(i+1, len(set(scid_list)),
                     status='Iterating through screens')

        assert i == 2



import pandas as pd
from idr import genes_of_interest_from_string
from idr import network_of_interest
from idr import get_entrezid
from idr import get_ensembleid
from idr import genes_of_interest_go
from idr import ensembleid_to_genesymbol
from IPython.display import Image

from .config import gene_symbol, taxonomy_id, ensemble_id


class TestExternalDBs():

    def test_genes_of_interest_from_string(self):
        intpartners = genes_of_interest_from_string(gene_symbol,
                                                    20,
                                                    taxonomy_id)
        assert isinstance(intpartners, pd.DataFrame)
        assert not intpartners.empty

    def test_network_of_interest(self):
        int_image = network_of_interest(gene_symbol,
                                        20)
        assert isinstance(int_image, Image)

    def test_get_entrezid(self):
        entrezid = get_entrezid(gene_symbol)
        assert isinstance(entrezid, list)
        assert (entrezid is not None)
        # could do a proper assert for the returned id, but
        # given the regular updates in ids this assert
        # will break very often
        # so sticking to basic asserts

    def test_get_ensembleid(self):
        ensembleid = get_ensembleid(gene_symbol)
        assert isinstance(ensembleid, list)
        assert (ensembleid is not None)

    def test_genes_of_interest_go(self):
        go_gene_list = genes_of_interest_go('GO:0005885', taxonomy_id)
        assert isinstance(go_gene_list, list)
        assert (go_gene_list != [])

    def test_ensembleid_to_genesymbol(self):

        gene_symbol = ensembleid_to_genesymbol(ensemble_id)

        assert isinstance(gene_symbol, str)
        assert (gene_symbol is not None)
